/*------------------------------------------------------------------------------
* readrnx.c : read rinex observation data/navigation message files
*
*          Copyright (C) 2006 by T.TAKASU, All rights reserved.
*
* version : $Revision: 1.2 $ $Date: 2006/12/12 23:27:39 $
* history : 2006/04/26 0.1 new
*-----------------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "rtkcmn.h"
#include "readrnx.h"

static const char rcsid[]="$Id: readrnx.c,v 1.2 2006/12/12 23:27:39 ttaka Exp ttaka $";

/* sort observation data -----------------------------------------------------*/
static int cmpobs(const void *p1, const void *p2)
{
	obsd_t *q1=(obsd_t *)p1,*q2=(obsd_t *)p2;
	double tt=timediff(q1->time,q2->time);
	
	if (ABS(tt)>MAXDT) return tt<0?-1:1;
	return q1->rcv!=q2->rcv?q1->rcv-q2->rcv:q1->sat-q2->sat;
}
static int sortobs(obs_t *obs)
{
	int i,j,n;
	
	if (obs->n<=0) return 0;
	qsort(obs->data,obs->n,sizeof(obsd_t),cmpobs);
	for (i=n=0;i<obs->n;i=j,n++) {
		for (j=i+1;j<obs->n;j++)
			if (timediff(obs->data[j].time,obs->data[i].time)>MAXDT) break;
	}
	return n;
}
/* delete duplicated navigation messages -------------------------------------*/
static int cmpnav(const void *p1, const void *p2)
{
	navm_t *q1=(navm_t *)p1,*q2=(navm_t *)p2;
	
	return q1->toe.time!=q2->toe.time?q1->toe.time-q2->toe.time:
		  (q1->toe.sec!=q2->toe.sec?q1->toe.sec-q2->toe.sec:q1->sat-q2->sat);
}
static void uniqnav(nav_t *nav)
{
	int i,j;
	
	if (nav->n<=0) return;
	qsort(nav->msg,nav->n,sizeof(navm_t),cmpnav);
	for (i=j=0;i<nav->n;i++) {
		if (nav->msg[i].sat!=nav->msg[j].sat||
			ABS(timediff(nav->msg[i].toe,nav->msg[j].toe))>0.001)
			nav->msg[++j]=nav->msg[i];
	}
	nav->n=j+1;
}
/* read rinex header ---------------------------------------------------------*/
static char readrnxh(FILE *fp, char tobs[][3], int *nt, double *ion)
{
	char buff[256],trnx=' ';
	int i,j,n;
	
	*nt=0;
	while (fgets(buff,sizeof(buff),fp)) {
		if (strlen(buff)<=60) continue;
		else if (strstr(buff+60,"RINEX VERSION / TYPE")) {
			trnx=*(buff+20);
		}
		else if (strstr(buff+60,"# / TYPES OF OBSERV")) {
			n=(int)str2num(buff,0,6);
			for (i=0,j=10;i<n;i++,j+=6,(*nt)++) {
				if (j>=58) {if (!fgets(buff,sizeof(buff),fp)) break; j=10;}
				strncpy(tobs[i],buff+j,2); tobs[i][2]='\0';
			}
		}
		else if (strstr(buff+60,"ION ALPHA")) {
			for (i=0,j=3;i<4;i++,j+=12) ion[i]=str2num(buff,j,11);
		}
		else if (strstr(buff+60,"ION BETA")) {
			for (i=0,j=3;i<4;i++,j+=12) ion[i+4]=str2num(buff,j,11);
		}
		else if (strstr(buff+60,"END OF HEADER")) break;
	}
	return trnx;
}
/* read rinex observation data -----------------------------------------------*/
static void readrnxo(FILE *fp, char tobs[][3], int nt, int rcv, obs_t *obs)
{
	const char *ts[]={"L1","L2","C1","P2"};
	char buff[256];
	short lock;
	int i,j,k,n,index[32]={0},sats[128];
	double value;
	gtime_t time;
	obsd_t *data;
	
	for (i=0;i<nt;i++)
		for (j=0;j<4;j++) if (!strcmp(tobs[i],ts[j])) index[i]=j+1;
	i=-1;
	while (fgets(buff,sizeof(buff),fp)) {
		if (i<0) {
			if (str2time(buff,0,26,&time)) continue;
			n=(int)str2num(buff,29,3);
			for (j=0,k=32;j<n;j++,k+=3) {
				if (k>=68) {if (!fgets(buff,sizeof(buff),fp)) break; k=32;}
				sats[j]=strchr(" G",*(buff+k))?(int)str2num(buff,k+1,2):0;
			}
		}
		else if (sats[i]>=0) {
			if (obs->siz<=obs->n) {
				obs->siz=obs->siz==0?32768:obs->siz*2;
				if (!(obs->data=realloc(obs->data,sizeof(obsd_t)*obs->siz))) {
					fprintf(stderr,"malloc error\n");
					obs->n=obs->siz=0; return;
				}
			}
			data=&obs->data[obs->n++];
			data->time=time; data->sat=sats[i]; data->rcv=rcv;
			for (j=0;j<2;j++) data->L[j]=data->P[j]=NAN;
			for (j=k=0;j<nt;j++,k+=16) {
				if (k>=80) {if (!fgets(buff,sizeof(buff),fp)) break; k=0;}
				value=str2num(buff,k,14); if (ABS(value)<1E-6) continue;
				lock=(short)str2num(buff,k+14,1)&1; /* 1:loss of lock */
				switch (index[j]) {
					case 1: data->L[0]=value; data->LLI[0]=lock; break;
					case 2: data->L[1]=value; data->LLI[1]=lock; break;
					case 3: data->P[0]=value; break;
					case 4: data->P[1]=value; break;
				}
			}
		}
		if (++i>=n) i=-1;
	}
}
/* read rinex navigation message ---------------------------------------------*/
static void readrnxn(FILE *fp, const double *ion, nav_t *nav)
{
	char buff[256];
	int i=0,j,sat;
	double e[64];
	gtime_t toc;
	navm_t *msg;
	
	while (fgets(buff,sizeof(buff),fp)) {
		if (i==0) {
			sat=(int)str2num(buff,0,2);
			if (str2time(buff,3,17,&toc)) continue;
		}
		for (j=i==0?22:3;j<=60;j+=19) e[i++]=str2num(buff,j,19);
		if (i>=31) {
			if (nav->siz<=nav->n) {
				nav->siz=nav->siz==0?1024:nav->siz*2;
				if (!(nav->msg=realloc(nav->msg,sizeof(navm_t)*nav->siz))) {
					fprintf(stderr,"malloc error\n");
					nav->n=nav->siz=0; return;
				}
			}
			msg=&nav->msg[nav->n++];
			msg->sat=sat; msg->iode=(int)e[3]; msg->iodc=(int)e[26];
			msg->sva=(int)e[23]; msg->svh=(int)e[24];
			msg->toe=gpst2time(e[21],e[11]); msg->toc=toc; 
			msg->A=SQR(e[10]); msg->e=e[ 8]; msg->i0  =e[15]; msg->OMG0=e[13];
			msg->omg =e[17]; msg->M0 =e[ 6]; msg->deln=e[ 5]; msg->OMGd=e[18];
			msg->idot=e[19]; msg->crc=e[16]; msg->crs =e[ 4]; msg->cuc =e[ 7];
			msg->cus =e[ 9]; msg->cic=e[12]; msg->cis =e[14]; msg->toes=e[11];
			msg->f0  =e[ 0]; msg->f1 =e[ 1]; msg->f2  =e[ 2]; msg->tgd =e[25];
			i=0;
		}
	}
	for (i=0;i<8;i++) nav->ion[i]=ion[i];
}
/* read rinex file -----------------------------------------------------------*/
static void readrnxf(FILE *fp, int rcv, obs_t *obs, nav_t *nav)
{
	int nt;
	char tobs[32][3];
	double ion[8]={ /* default ionosphere model parameters */
		0.1676E-7,-0.7451E-8,-0.5960E-7,0.1192E-6,1.352E5,-1.802E5,6.554E4,0.0
	};
	switch (readrnxh(fp,tobs,&nt,ion)) {
		case 'O': readrnxo(fp,tobs,nt,rcv,obs); break;
		case 'N': readrnxn(fp,ion,nav); break;
	}
}
/* read rinex files ------------------------------------------------------------
* read rinex observation data(obs)/navigation message(nav) files
* args   : char **files	 I		rinex files
*		   int	 n		 I		number of rinex files (0:input from stdin)
*		   obs_t *obs	 O		observation data
*		   nav_t *nav	 O		navigation messages
* return : number of epochs (0:no data)
* note   : obs data are sorted by time, receiver and satellite number
*          RINEX 2.10, only support gps.
*-----------------------------------------------------------------------------*/
extern int readrnx(char **files, int n, obs_t *obs, nav_t *nav)
{
	FILE *fp;
	int i,rcv=0;
	
	trace_("readrnxn()\n");
	
	obs->n=obs->siz=nav->n=nav->siz=0; obs->data=NULL; nav->msg=NULL;
	
	if (n>0) {
		for (i=0;i<n;i++) {
			if (!(fp=fopen(files[i],"r"))) {
				fprintf(stderr,"file open error : %s\n",files[i]);
				continue;
			}
			readrnxf(fp,++rcv,obs,nav);
			fclose(fp);
		}
	}
	else readrnxf(stdin,1,obs,nav);
	
	uniqnav(nav);
	return sortobs(obs);
}
