/*------------------------------------------------------------------------------
* rnx2pntp.c : read rinex obs/nav and compute receiver positions
*
* synopsis :
*      rnx2pntp [option]... file file [...]
*
* description :
*      read rinex obs/nav files and output receiver positions by single point
*      positioning. options are as follows. ([]:default)
*
*      -o output output file [stdout]
*      -e elmin  elevation cutoff angle (deg) [10]
*      -h        output header [no]
*      -s sep    field separator [' ']
*      -t        yyyy/mm/dd hh:mm:ss as time format [gpstime]
*      -u        columns after decimal point of time [0]
*      -x        ecef x/y/z(m) as receiver position [lat/lon/h(deg,m)]
*
*          Copyright (C) 2006 by T.TAKASU, All rights reserved.
*
* version : $Revision: 1.1 $ $Date: 2006/12/12 23:27:39 $
* history : 2006/12/09  0.1 new
*-----------------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "rtkcmn.h"
#include "readrnx.h"
#include "pntpos.h"

static const char rcsid[]="$Id: rnx2pntp.c,v 1.1 2006/12/12 23:27:39 ttaka Exp ttaka $";

#define MAXFILE		16			/* max number of input files */

static FILE *fp;				/* output file pointer */
static double elmin=10.0;		/* elevation cutoff angle (deg) */
static int headf=0;				/* output header flag (0:off,1:on) */
static int timef=0;				/* time format (0:gpstime,1:hh:mm:ss) */
static int timeu=0;				/* columns after decimal point */
static int xyzf=0;				/* position flag (0:lat/lon/h,1:xyz) */
static int iobs=0;				/* observation data record index */
static char sep=' ';			/* field separator */
static obs_t obss;				/* observation data */
static nav_t navs;				/* navigation messages */

/* input observation data/navigation messages --------------------------------*/
static int input(obsd_t **obs, nav_t **nav)
{
	int i;
	
	for (i=iobs;iobs<obss.n;iobs++) {
		if (timediff(obss.data[iobs].time,obss.data[i].time)>0.0) break;
	}
	*obs=obss.data+i; *nav=&navs;
	return iobs-i;
}
/* output header ------------------------------------------------------------*/
static void outhead(FILE *fp, char **file, int n)
{
	int i,w1,w2;
	char s1[32],s2[32];
	double t1,t2;
	
	fprintf(fp,"%% INPUT FILES   : ");
	for (i=0;i<n;i++) fprintf(fp,"%s ",file[i]);
	fprintf(fp,"\n");
	time2str(obss.data[0].time,s1,3);
	time2str(obss.data[obss.n-1].time,s2,3);
	t1=time2gpst(obss.data[0].time,&w1);
	t2=time2gpst(obss.data[obss.n-1].time,&w2);
	fprintf(fp,"%% OBS START     : %s (WEEK%04d %9.3f)\n",s1,w1,t1);
	fprintf(fp,"%% OBS END       : %s (WEEK%04d %9.3f)\n",s2,w2,t2);
	fprintf(fp,"%%\n%%  %-*s",(timef?17:5)+(timeu<=0?0:timeu+1),"TIME");
	if (xyzf) fprintf(fp,"%14s %14s %14s\n","ECEF-X(M)","ECEF-Y(M)","ECEF-Z(M)");
	else      fprintf(fp,"%14s %14s %10s\n","LATI(DEG)","LONGI(DEG)","HGT(M)");
}
/* output result -------------------------------------------------------------*/
static void output(gtime_t t, double *rr)
{
	int week;
	char s[32];
	double pos[3];
	
	if (timef) time2str(t,s,timeu);
	else sprintf(s,"%*.*f",7+(timeu<=0?0:timeu+1),timeu,time2gpst(t,&week));

	if (xyzf) {
		fprintf(fp,"%s%c%14.4f%c%14.4f%c%14.4f\n",s,sep,rr[0],sep,rr[1],sep,
			    rr[2]);
	}
	else {
		ecef2pos(rr,pos);
		fprintf(fp,"%s%c%14.9f%c%14.9f%c%10.4f\n",s,sep,pos[0]*R2D,sep,
			    pos[1]*R2D,sep,pos[2]);
	}
}
/* rnx2pos main --------------------------------------------------------------*/
int main(int argc, char **argv)
{
	int i,n;
	char *infile[MAXFILE],*outfile="";
	
	/* process options */
	for (i=1,n=0;i<argc;i++) {
		if      (!strcmp(argv[i],"-o")&&i+1<argc) outfile=argv[++i];
		else if (!strcmp(argv[i],"-e")&&i+1<argc) elmin=atof(argv[++i]);
		else if (!strcmp(argv[i],"-s")&&i+1<argc) sep=*argv[++i];
		else if (!strcmp(argv[i],"-u")&&i+1<argc) timeu=atoi(argv[++i]);
		else if (!strcmp(argv[i],"-h")) headf=1;
		else if (!strcmp(argv[i],"-t")) timef=1;
		else if (!strcmp(argv[i],"-x")) xyzf=1;
		else if (n<MAXFILE) infile[n++]=argv[i];
	}
	if (n<=0) {
		fprintf(stderr,"no input file\n");
		return -1;
	}
	/* read rinex obs/nav files */
	if (readrnx(infile,n,&obss,&navs)<=0) {
		fprintf(stderr,"no observation data\n");
		return -2;
	}
	if (navs.n<=0) {
		fprintf(stderr,"no navigation message\n");
		return -3;
	}
	if (*outfile&&!(fp=fopen(outfile,"w"))) {
		fprintf(stderr,"output file open error : %s\n",outfile);
		return -4;
	}
	else fp=stdout;
	
	/* output header */
	if (headf) outhead(fp,infile,n);
	
	/* point positiong */
	pntpos(input,output,elmin);
	
	fclose(fp);
	return 0;
}
