/*------------------------------------------------------------------------------
* rnx2rtkp.c : compute receiver positions by rtk precise positioning
*
* synopsis :
*      rinex2rtkp [option]... file file [...]
*
*          Copyright (C) 2006 by T.TAKASU, All rights reserved.
*
* version : $Revision: 1.1 $ $Date: 2006/12/12 23:27:39 $
* history : 2006/12/09  0.1 new
*-----------------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "rtkcmn.h"
#include "readrnx.h"
#include "pntpos.h"
#include "rtkpos.h"

static const char rcsid[]="$Id: rnx2rtkp.c,v 1.1 2006/12/12 23:27:39 ttaka Exp ttaka $";

#define MAXFILE		4			/* max number of input files */

static FILE *fp;				/* output file pointer */
static double elmin=10.0;		/* elevation cutoff angle (deg) */
static int mode=1;				/* positioning mode (0:static,1:kinematic) */
static int headf=0;				/* output header flag (0:off,1:on) */
static int timef=0;				/* time format (0:gpstime,1:hh:mm:ss) */
static int timeu=0;				/* columns after decimal point */
static int xyzf=0;				/* position flag (0:lat/lon/h,1:xyz) */
static int iobs=0;				/* observation data record index */
static char sep=' ';			/* field separator */
static double rb[3]={0.0};		/* reference station position (ecef) (m) */
static obs_t obss;				/* observation data */
static nav_t navs;				/* navigation messages */

/* input observation data/navigation messages --------------------------------*/
static int input(obsd_t **obs, nav_t **nav)
{
	int i;
	
	for (i=iobs;iobs<obss.n;iobs++) {
		if (timediff(obss.data[iobs].time,obss.data[i].time)>MAXDT) break;
	}
	*obs=obss.data+i; *nav=&navs;
	return iobs-i;
}
/* output header ------------------------------------------------------------*/
static void outhead(FILE *fp, char **file, int n)
{
	int i,w1,w2;
	char s1[32],s2[32],*s3="STATIC",*s4="KINEMATC";
	double t1,t2,pos[3];
	
	fprintf(fp,"%% INPUT FILES   : ");
	for (i=0;i<n;i++) fprintf(fp,"%s ",file[i]);
	fprintf(fp,"\n");
	time2str(obss.data[0].time,s1,3);
	time2str(obss.data[obss.n-1].time,s2,3);
	t1=time2gpst(obss.data[0].time,&w1);
	t2=time2gpst(obss.data[obss.n-1].time,&w2);
	fprintf(fp,"%% OBS START     : %s (WEEK%04d %9.3f)\n",s1,w1,t1);
	fprintf(fp,"%% OBS END       : %s (WEEK%04d %9.3f)\n",s2,w2,t2);
	fprintf(fp,"%% POS MODE      : %s\n",mode?s4:s3);
	fprintf(fp,"%% REFERENCE POS :");
	if (xyzf)
		fprintf(fp,"%14.4f%c%14.4f%c%14.4f\n",rb[0],sep,rb[1],sep,rb[2]);
	else {
		ecef2pos(rb,pos);
		fprintf(fp,"%14.9f%c%14.9f%c%10.4f\n",pos[0]*R2D,sep,pos[1]*R2D,sep,pos[2]);
	}
	fprintf(fp,"%%\n%%  %-*s",(timef?17:5)+(timeu<=0?0:timeu+1),"TIME");
	if (xyzf) {
		fprintf(fp,"%14s %14s %14s %7s %7s %7s %3s\n","ECEF-X(M)","ECEF-Y(M)",
			    "ECEF-Z(M)","SDX(M)","SDY(M)","SDZ(M)","Q");
	}
	else {
		fprintf(fp,"%14s %14s %10s %3s\n","LATI(DEG)","LONGI(DEG)","HGT(M)","Q");
	}
}
/* output result -------------------------------------------------------------*/
static void output(gtime_t t, double *rr)
{
	int week;
	char s[32];
	double pos[3];
	
	if (timef) time2str(t,s,timeu);
	else sprintf(s,"%*.*f",7+(timeu<=0?0:timeu+1),timeu,time2gpst(t,&week));

	if (xyzf) {
		fprintf(fp,"%s%c%14.4f%c%14.4f%c%14.4f%c%7.4f%c%7.4f%c%7.4f%c%3.0f\n",
				s,sep,rr[0],sep,rr[1],sep,rr[2],sep,rr[3],sep,rr[4],sep,rr[5],
				sep,rr[6]);
	}
	else {
		ecef2pos(rr,pos);
		fprintf(fp,"%s%c%14.9f%c%14.9f%c%10.4f%c%3.0f\n",s,sep,pos[0]*R2D,sep,
			    pos[1]*R2D,sep,pos[2],sep,rr[6]);
	}
}
/* compute reference station position as mean position -----------------------*/
static int refpos(double *rb)
{
	int i,j,k,n;
	double rs[3*MAXOBS],dts[MAXOBS],azel[2*MAXOBS],rr[3],dtr;
	
	for (i=n=0;i<obss.n;i=j) {
		for (;i<obss.n;i++) if (obss.data[i].rcv==2) break;
		for (j=i;j<obss.n;j++) {
			if (timediff(obss.data[j].time,obss.data[i].time)>0.0) break;
		}
		satpos(obss.data+i,j-i,&navs,rs,dts);
		
		if (pointp(obss.data+i,j-i,rs,dts,navs.ion,10.0,rr,&dtr,azel)) continue;
		
		for (k=0;k<3;k++) rb[k]+=rr[k]; n++;
	}
	if (n<=0) return -1; else for (k=0;k<3;k++) rb[k]/=n;
	return 0;
}
/* rnx2rtkp main -------------------------------------------------------------*/
int main(int argc, char **argv)
{
	int i,j,n;
	char *infile[MAXFILE],*outfile="";
	
	/* process options */
	for (i=1,n=0;i<argc;i++) {
		if      (!strcmp(argv[i],"-o")&&i+1<argc) outfile=argv[++i];
		else if (!strcmp(argv[i],"-m")&&i+1<argc) mode=0;
		else if (!strcmp(argv[i],"-e")&&i+1<argc) elmin=atof(argv[++i]);
		else if (!strcmp(argv[i],"-s")&&i+1<argc) sep=*argv[++i];
		else if (!strcmp(argv[i],"-u")&&i+1<argc) timeu=atoi(argv[++i]);
		else if (!strcmp(argv[i],"-h")) headf=1;
		else if (!strcmp(argv[i],"-t")) timef=1;
		else if (!strcmp(argv[i],"-x")) xyzf=1;
		else if (!strcmp(argv[i],"-r")&&i+3<argc)
			for (j=0;j<3;j++) rb[j]=atof(argv[++i]);
		else if (n<MAXFILE) infile[n++]=argv[i];
	}
	if (n<=0) {
		fprintf(stderr,"no input file\n");
		return -1;
	}
	/* read rinex obs/nav files */
	if (readrnx(infile,n,&obss,&navs)<=0) {
		fprintf(stderr,"no observation data\n");
		return -2;
	}
	if (navs.n<=0) {
		fprintf(stderr,"no navigation message\n");
		return -3;
	}
	if (*outfile&&!(fp=fopen(outfile,"w"))) {
		fprintf(stderr,"output file open error : %s\n",outfile);
		return -4;
	}
	else fp=stdout;
	
	if (NORM(rb)<=0.0&&refpos(rb)) {
		fprintf(stderr,"reference station position error\n");
		return -5;
	}
	/* output header */
	if (headf) outhead(fp,infile,n);
	
	/* rtk positiong */
	rtkpos(input,output,mode,rb,elmin);
	
	fclose(fp);
	return 0;
}
